import logging
import time

from PIL import Image, ImageChops
from PIL.ImageGrab import grab

from pyvirtualdisplay import Display

log = logging.getLogger(__name__)


class DisplayTimeoutError(Exception):
    pass


def autocrop(im, bgcolor):
    """Crop borders off an image.

    :param im: Source image.
    :param bgcolor: Background color, using either a color tuple.
    :return: An image without borders, or None if there's no actual content in the image.
    """
    if im.mode != "RGB":
        im = im.convert("RGB")
    bg = Image.new("RGB", im.size, bgcolor)
    diff = ImageChops.difference(im, bg)
    bbox = diff.getbbox()
    if bbox:
        return im.crop(bbox)
    return None  # no contents


class SmartDisplay(Display):
    def autocrop(self, im):
        """Crop borders off an image.

        :param im: Source image.
        :return: An image without borders, or None if there's no actual content in the image.
        """
        return autocrop(im, self._bgcolor)

    def grab(self, autocrop=True):
        # TODO: use Xvfb fbdir option for screenshot
        img = grab(xdisplay=self.new_display_var)

        if autocrop:
            img = self.autocrop(img)
        return img

    def waitgrab(self, timeout=60, autocrop=True, cb_imgcheck=None):
        """start process and create screenshot.
        Repeat screenshot until it is not empty and
        cb_imgcheck callback function returns True
        for current screenshot.

        :param autocrop: True -> crop screenshot
        :param timeout: int
        :param cb_imgcheck: None or callback for testing img,
                            True = accept img,
                            False = reject img
        """
        t = 0
        sleep_time = 0.3  # for fast windows
        repeat_time = 0.5
        while 1:
            log.debug("sleeping %s secs" % str(sleep_time))
            time.sleep(sleep_time)
            t += sleep_time
            img = self.grab(autocrop=False)
            img_crop = self.autocrop(img)
            if autocrop:
                img = img_crop
            if img_crop:
                if not cb_imgcheck:
                    break
                if cb_imgcheck(img):
                    break
            sleep_time = repeat_time
            repeat_time += 0.5  # progressive
            if t > timeout:
                msg = "Timeout! elapsed time:%s timeout:%s " % (t, timeout)
                raise DisplayTimeoutError(msg)
                # break

            log.debug("screenshot is empty, next try..")
        assert img
        #        if not img:
        #            log.debug('screenshot is empty!')
        return img



import logging

from pyvirtualdisplay.about import __version__
from pyvirtualdisplay.display import Display

Display  # ignore unused

log = logging.getLogger(__name__)

log = logging.getLogger(__name__)
log.debug("version=%s", __version__)



"Nested Xephyr servers"
from easyprocess import EasyProcess

from pyvirtualdisplay import Display

with Display(visible=True, size=(220, 180), bgcolor="black"):
    with Display(visible=True, size=(200, 160), bgcolor="white"):
        with Display(visible=True, size=(180, 140), bgcolor="black"):
            with Display(visible=True, size=(160, 120), bgcolor="white"):
                with Display(visible=True, size=(140, 100), bgcolor="black"):
                    with Display(visible=True, size=(120, 80), bgcolor="white"):
                        with Display(visible=True, size=(100, 60), bgcolor="black"):
                            with EasyProcess(["xmessage", "hello"]) as proc:
                                proc.wait()



import logging

from pyvirtualdisplay.abstractdisplay import AbstractDisplay

log = logging.getLogger(__name__)

PROGRAM = "Xephyr"


class XephyrDisplay(AbstractDisplay):
    """
    Xephyr wrapper

    Xephyr is an X server outputting to a window on a pre-existing X display
    """

    def __init__(
        self,
        size=(1024, 768),
        color_depth=24,
        bgcolor="black",
        use_xauth=False,
        retries=10,
        extra_args=[],
        manage_global_env=True,
        parent=None,
    ):
        """
        :param bgcolor: 'black' or 'white'
        """
        self._color_depth = color_depth
        self._size = size
        self._bgcolor = bgcolor
        self._parent = parent

        AbstractDisplay.__init__(
            self,
            PROGRAM,
            use_xauth=use_xauth,
            retries=retries,
            extra_args=extra_args,
            manage_global_env=manage_global_env,
        )

    def _check_flags(self, helptext):
        self._has_resizeable = "-resizeable" in helptext

    def _cmd(self):
        cmd = (
            [
                PROGRAM,
            ]
            + (["-parent", self._parent] if self._parent else [])
            + [
                dict(black="-br", white="-wr")[self._bgcolor],
                "-screen",
                "x".join(map(str, list(self._size) + [self._color_depth])),
            ]
        )
        if self._has_displayfd:
            cmd += ["-displayfd", str(self._pipe_wfd)]
        else:
            cmd += [self.new_display_var]

        if self._has_resizeable:
            cmd += ["-resizeable"]
        return cmd



"""Utility functions for xauth."""
import hashlib
import os
import subprocess


class NotFoundError(Exception):
    """Error when xauth was not found."""


def is_installed():
    """
    Return whether or not xauth is installed.
    """
    try:
        xauth = subprocess.Popen(
            ["xauth", "-V"],
            # env=self._env(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        _, _ = xauth.communicate()
        # p = EasyProcess(["xauth", "-V"])
        # p.enable_stdout_log = False
        # p.enable_stderr_log = False
        # p.call()
    except FileNotFoundError:
        return False
    else:
        return True


def generate_mcookie():
    """
    Generate a cookie string suitable for xauth.
    """
    data = os.urandom(16)  # 16 bytes = 128 bit
    return hashlib.md5(data).hexdigest()


def call(*args):
    """
    Call xauth with the given args.
    """
    xauth = subprocess.Popen(
        ["xauth"] + list(args),
        # env=self._env(),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    _, _ = xauth.communicate()
    # EasyProcess(["xauth"] + list(args)).call()



from easyprocess import EasyProcess

from pyvirtualdisplay.smartdisplay import SmartDisplay


def test_double():
    with SmartDisplay(visible=False, bgcolor="black") as disp:
        with EasyProcess(["xmessage", "hello1"]):
            img = disp.waitgrab()
            assert img is not None

    with SmartDisplay(visible=False, bgcolor="black") as disp:
        with EasyProcess(["xmessage", "hello2"]):
            img = disp.waitgrab()
            assert img is not None



import logging
from threading import Thread
from time import sleep

from easyprocess import EasyProcess
from PIL import ImageChops

from pyvirtualdisplay.smartdisplay import SmartDisplay


def func2(results):
    with SmartDisplay(manage_global_env=False) as disp:
        env = disp.env()
        logging.info("env=%s", env)
        sleep(2)
        with EasyProcess(["xmessage", "hello"], env=env):
            im = disp.waitgrab(timeout=1)
            results[0] = im
            sleep(2)


def test_smart():
    results = [None]
    t = Thread(target=func2, args=(results,))
    t.start()
    sleep(1)
    with SmartDisplay(manage_global_env=False) as disp:
        env = disp.env()
        logging.info("env=%s", env)
        with EasyProcess(["xmessage", "hello"], env=env):
            sleep(2)
            im0 = disp.waitgrab(timeout=1)
            assert im0
    t.join()
    im1 = results[0]
    assert im1
    # im0.save('/vagrant/im0.png')
    # im1.save('/vagrant/im1.png')
    img_diff = ImageChops.difference(im0, im1)
    ex = img_diff.getextrema()
    logging.debug("diff getextrema: %s", ex)
    diff_bbox = img_diff.getbbox()
    assert diff_bbox is None



from tutil import has_xvnc, rfbport

from pyvirtualdisplay import Display


def test_with_xvfb():
    with Display(size=(800, 600)) as vd:
        assert vd.is_alive()
        assert vd._backend == "xvfb"
    # assert vd.return_code == 0
    assert not vd.is_alive()

    with Display(visible=False, size=(800, 600)) as vd:
        assert vd.is_alive()
        assert vd._backend == "xvfb"
    # assert vd.return_code == 0
    assert not vd.is_alive()

    with Display(backend="xvfb", size=(800, 600)) as vd:
        assert vd.is_alive()
        assert vd._backend == "xvfb"
    # assert vd.return_code == 0
    assert not vd.is_alive()


def test_with_xephyr():
    with Display() as vd:
        with Display(visible=True, size=(800, 600)) as vd:
            assert vd.is_alive()
            assert vd._backend == "xephyr"
        # assert vd.return_code == 0
        assert not vd.is_alive()

        with Display(backend="xephyr", size=(800, 600)) as vd:
            assert vd.is_alive()
            assert vd._backend == "xephyr"
        # assert vd.return_code == 0
        assert not vd.is_alive()


if has_xvnc():

    def test_with_xvnc():
        with Display(backend="xvnc", size=(800, 600), rfbport=rfbport()) as vd:
            assert vd.is_alive()
            assert vd._backend == "xvnc"
        # assert vd.return_code == 0
        assert not vd.is_alive()


def test_dpi():
    with Display(backend="xvfb", size=(800, 600), dpi=99) as vd:
        assert vd.is_alive()
    # assert vd.return_code == 0
    assert not vd.is_alive()



import os

from tutil import prog_check

from pyvirtualdisplay import Display, xauth

if prog_check(["xauth", "-V"]):

    def test_xauth_is_installed():
        assert xauth.is_installed()

    def test_xauth():
        """
        Test that a Xauthority file is created.
        """
        old_xauth = os.getenv("XAUTHORITY")
        display = Display(visible=False, use_xauth=True)
        display.start()
        new_xauth = os.getenv("XAUTHORITY")

        assert new_xauth is not None
        assert os.path.isfile(new_xauth)
        filename = os.path.basename(new_xauth)
        assert filename.startswith("PyVirtualDisplay.")
        assert filename.endswith("Xauthority")

        display.stop()
        assert old_xauth == os.getenv("XAUTHORITY")
        assert not os.path.isfile(new_xauth)



"Start Xvfb server and open xmessage window. Thread safe."

import threading

from easyprocess import EasyProcess

from pyvirtualdisplay.smartdisplay import SmartDisplay


def thread_function(index):
    # manage_global_env=False is thread safe
    with SmartDisplay(manage_global_env=False) as disp:
        cmd = ["xmessage", str(index)]
        # disp.new_display_var should be used for new processes
        # disp.env() copies global os.environ and adds disp.new_display_var
        with EasyProcess(cmd, env=disp.env()):
            img = disp.waitgrab()
            img.save("xmessage{}.png".format(index))


t1 = threading.Thread(target=thread_function, args=(1,))
t2 = threading.Thread(target=thread_function, args=(2,))
t1.start()
t2.start()
t1.join()
t2.join()



import sys
from time import sleep

from easyprocess import EasyProcess
from entrypoint2 import entrypoint
from tutil import has_xvnc, worker

from pyvirtualdisplay import Display

# ubuntu 14.04 no displayfd
# ubuntu 16.04 displayfd

# Xephyr leaks shared memory segments until 18.04
# 		https://gitlab.freedesktop.org/xorg/xserver/-/issues/130
#       approximately 4MB/call


def test_race_100_xvfb():
    check_n(100, "xvfb")


# TODO: this fails with "Xephyr cannot open host display"
#   Xephyr bug?
# def test_race_500_100_xephyr():
#     for _ in range(500):
#         check_n(100, "xephyr")


if has_xvnc():

    def test_race_10_xvnc():
        check_n(10, "xvnc")


def check_n(n, backend):
    with Display():
        ls = []
        try:
            for i in range(n):
                cmd = [
                    sys.executable,
                    __file__.rsplit(".", 1)[0] + ".py",
                    str(i),
                    backend,
                    str(n),
                    "--debug",
                ]
                p = EasyProcess(cmd)
                p.start()
                ls += [p]

            sleep(3)

            good_count = 0
            rc_ls = []
            for p in ls:
                p.wait()
                if p.return_code == 0:
                    good_count += 1
                rc_ls += [p.return_code]
        finally:
            for p in ls:
                p.stop()
        print(rc_ls)
        print(good_count)
        assert good_count == n


@entrypoint
def main(i, backend, retries):
    retries = int(retries)
    kwargs = dict()
    if backend == "xvnc":
        kwargs["rfbport"] = 42000 + 100 * worker() + int(i)
    # print("$DISPLAY=%s" % (os.environ.get("DISPLAY")))
    d = Display(backend=backend, retries=retries, **kwargs).start()
    print(
        "my index:%s  backend:%s disp:%s retries:%s"
        % (
            i,
            backend,
            d.new_display_var,
            d._obj._retries_current,
        )
    )
    ok = d.is_alive()
    d.stop()
    assert ok



import logging
import os
import sys
from tempfile import TemporaryDirectory
from time import sleep

from easyprocess import EasyProcess
from tutil import has_xvnc, kill_process_tree, prog_check, worker

from pyvirtualdisplay import Display

log = logging.getLogger(__name__)


python = sys.executable


def test_screenshot():
    owd = os.getcwd()
    with TemporaryDirectory(prefix="pyvirtualdisplay_") as tmpdirname:
        try:
            os.chdir(tmpdirname)
            p = EasyProcess([python, "-m", "pyvirtualdisplay.examples.screenshot"])
            p.call()
            assert p.return_code == 0
        finally:
            os.chdir(owd)


def test_headless():
    p = EasyProcess([python, "-m", "pyvirtualdisplay.examples.headless"]).start()
    sleep(1)
    assert p.is_alive()
    # p.stop()
    kill_process_tree(p)


def test_nested():
    with Display():
        p = EasyProcess([python, "-m", "pyvirtualdisplay.examples.nested"]).start()
        sleep(1)
        assert p.is_alive()
        # p.stop()
        kill_process_tree(p)


if has_xvnc():

    def test_vncserver():
        if worker() == 0:
            p = EasyProcess(
                [python, "-m", "pyvirtualdisplay.examples.vncserver"]
            ).start()
            sleep(1)
            assert p.is_alive()
            # p.stop()
            kill_process_tree(p)


if prog_check(["gnumeric", "-help"]):

    def test_lowres():
        with Display():
            p = EasyProcess([python, "-m", "pyvirtualdisplay.examples.lowres"]).start()
            sleep(1)
            assert p.is_alive()
            # p.stop()
            kill_process_tree(p)



import logging
import os

import psutil
from easyprocess import EasyProcess

from pyvirtualdisplay.util import get_helptext

log = logging.getLogger(__name__)


def prog_check(cmd):
    try:
        p = EasyProcess(cmd)
        p.enable_stdout_log = False
        p.enable_stderr_log = False
        p.call()
        return p.return_code == 0
    except Exception:
        return False


# def platform_is_osx():
#     return sys.platform == "darwin"


def has_displayfd():
    return "-displayfd" in get_helptext("Xvfb")


def has_xvnc():
    return prog_check(["Xvnc", "-help"])


def worker():
    w = 0
    PYTEST_XDIST_WORKER = os.environ.get("PYTEST_XDIST_WORKER")
    if PYTEST_XDIST_WORKER:
        # gw42
        w = int(PYTEST_XDIST_WORKER[2:])
    return w


def rfbport():
    port = 5900 + worker() + 9876
    log.info("rfbport=%s", port)
    return port


def kill_process_tree(ep):
    parent_pid = ep.pid
    parent = psutil.Process(parent_pid)
    for child in parent.children(recursive=True):
        try:
            child.kill()
        except psutil.NoSuchProcess:
            log.warning("NoSuchProcess error in kill_process_tree")
    parent.kill()



import fnmatch
import logging
import os
import select
import subprocess
import tempfile
import time
from threading import Lock

from pyvirtualdisplay import xauth
from pyvirtualdisplay.util import get_helptext, platform_is_osx

log = logging.getLogger(__name__)

# try:
#     import fcntl
# except ImportError:
#     fcntl = None

_mutex = Lock()
_mutex_popen = Lock()


_MIN_DISPLAY_NR = 1000
_USED_DISPLAY_NR_LIST = []

_X_START_TIMEOUT = 10
_X_START_TIME_STEP = 0.1
_X_START_WAIT = 0.1


class XStartTimeoutError(Exception):
    pass


class XStartError(Exception):
    pass


def _lock_files():
    tmpdir = "/tmp"
    try:
        ls = os.listdir(tmpdir)
    except FileNotFoundError:
        log.warning("missing /tmp")
        return []
    pattern = ".X*-lock"
    names = fnmatch.filter(ls, pattern)
    ls = [os.path.join(tmpdir, child) for child in names]
    ls = [p for p in ls if os.path.isfile(p)]
    return ls


def _search_for_display():
    # search for free display
    ls = list(map(lambda x: int(x.split("X")[1].split("-")[0]), _lock_files()))
    if len(ls):
        display = max(_MIN_DISPLAY_NR, max(ls) + 3)
    else:
        display = _MIN_DISPLAY_NR

    return display


class AbstractDisplay(object):
    """
    Common parent for X servers (Xvfb,Xephyr,Xvnc)
    """

    def __init__(self, program, use_xauth, retries, extra_args, manage_global_env):
        self._extra_args = extra_args
        self._retries = retries
        self._program = program
        self.stdout = None
        self.stderr = None
        self.old_display_var = None
        self._subproc = None
        self.display = None
        self._is_started = False
        self._manage_global_env = manage_global_env
        self._reset_global_env = False
        self._pipe_wfd = None
        self._retries_current = 0

        helptext = get_helptext(program)
        self._has_displayfd = "-displayfd" in helptext
        if not self._has_displayfd:
            log.debug("-displayfd flag is missing.")
        PYVIRTUALDISPLAY_DISPLAYFD = os.environ.get("PYVIRTUALDISPLAY_DISPLAYFD")
        if PYVIRTUALDISPLAY_DISPLAYFD:
            log.debug("PYVIRTUALDISPLAY_DISPLAYFD=%s", PYVIRTUALDISPLAY_DISPLAYFD)
            # '0'->false, '1'->true
            self._has_displayfd = bool(int(PYVIRTUALDISPLAY_DISPLAYFD))
        else:
            # TODO: macos: displayfd is available on XQuartz-2.7.11 but it doesn't work, always 0 is returned
            if platform_is_osx():
                self._has_displayfd = False

        self._check_flags(helptext)

        if use_xauth and not xauth.is_installed():
            raise xauth.NotFoundError()

        self._use_xauth = use_xauth
        self._old_xauth = None
        self._xauth_filename = None

    def _check_flags(self, helptext):
        pass

    def _cmd(self):
        raise NotImplementedError()

    def _redirect_display(self, on):
        """
        on:
         * True -> set $DISPLAY to virtual screen
         * False -> set $DISPLAY to original screen

        :param on: bool
        """
        d = self.new_display_var if on else self.old_display_var
        if d is None:
            log.debug("unset $DISPLAY")
            try:
                del os.environ["DISPLAY"]
            except KeyError:
                log.warning("$DISPLAY was already unset.")
        else:
            log.debug("set $DISPLAY=%s", d)
            os.environ["DISPLAY"] = d

    def _env(self):
        env = os.environ.copy()
        env["DISPLAY"] = self.new_display_var
        return env

    def start(self):
        """
        start display

        :rtype: self
        """
        if self._is_started:
            raise XStartError(self, "Display was started twice.")
        self._is_started = True

        if self._has_displayfd:
            self._start1_has_displayfd()
        else:
            i = 0
            while True:
                self._retries_current = i + 1
                try:
                    self._start1()
                    break
                except XStartError:
                    log.warning("start failed %s", i + 1)
                    time.sleep(0.05)
                    i += 1
                    if i >= self._retries:
                        raise XStartError(
                            "No success after %s retries. Last stderr: %s"
                            % (self._retries, self.stderr)
                        )
        if self._manage_global_env:
            self._redirect_display(True)
            self._reset_global_env = True

    def _popen(self, use_pass_fds):
        with _mutex_popen:
            if use_pass_fds:
                self._subproc = subprocess.Popen(
                    self._command,
                    pass_fds=[self._pipe_wfd],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=False,
                )
            else:
                self._subproc = subprocess.Popen(
                    self._command,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=False,
                )

    def _start1_has_displayfd(self):
        # stdout doesn't work on osx -> create own pipe
        rfd, self._pipe_wfd = os.pipe()

        self._command = self._cmd() + self._extra_args
        log.debug("command: %s", self._command)

        self._popen(use_pass_fds=True)

        self.display = int(self._wait_for_pipe_text(rfd))
        os.close(rfd)
        os.close(self._pipe_wfd)

        self.new_display_var = ":%s" % int(self.display)

        if self._use_xauth:
            self._setup_xauth()

        # https://github.com/ponty/PyVirtualDisplay/issues/2
        # https://github.com/ponty/PyVirtualDisplay/issues/14
        self.old_display_var = os.environ.get("DISPLAY", None)

    def _start1(self):
        with _mutex:
            self.display = _search_for_display()
            while self.display in _USED_DISPLAY_NR_LIST:
                self.display += 1
            self.new_display_var = ":%s" % int(self.display)

            _USED_DISPLAY_NR_LIST.append(self.display)

        self._command = self._cmd() + self._extra_args
        log.debug("command: %s", self._command)

        self._popen(use_pass_fds=False)

        self.new_display_var = ":%s" % int(self.display)

        if self._use_xauth:
            self._setup_xauth()

        # https://github.com/ponty/PyVirtualDisplay/issues/2
        # https://github.com/ponty/PyVirtualDisplay/issues/14
        self.old_display_var = os.environ.get("DISPLAY", None)

        # wait until X server is active
        start_time = time.time()

        d = self.new_display_var
        ok = False
        time.sleep(0.05)  # give time for early exit
        while True:
            if not self.is_alive():
                break

            try:
                xdpyinfo = subprocess.Popen(
                    ["xdpyinfo"],
                    env=self._env(),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=False,
                )
                _, _ = xdpyinfo.communicate()
                exit_code = xdpyinfo.returncode
            except FileNotFoundError:
                log.warning(
                    "xdpyinfo was not found, X start can not be checked! Please install xdpyinfo!"
                )
                time.sleep(_X_START_WAIT)  # old method
                ok = True
                break
            # try:
            #     xdpyinfo = EasyProcess(["xdpyinfo"], env=self._env())
            #     xdpyinfo.enable_stdout_log = False
            #     xdpyinfo.enable_stderr_log = False
            #     exit_code = xdpyinfo.call().return_code
            # except EasyProcessError:
            #     log.warning(
            #         "xdpyinfo was not found, X start can not be checked! Please install xdpyinfo!"
            #     )
            #     time.sleep(_X_START_WAIT)  # old method
            #     ok = True
            #     break

            if exit_code != 0:
                pass
            else:
                log.info('Successfully started X with display "%s".', d)
                ok = True
                break

            if time.time() - start_time >= _X_START_TIMEOUT:
                break
            time.sleep(_X_START_TIME_STEP)
        if not self.is_alive():
            log.warning("process exited early. stderr:%s", self.stderr)
            msg = "Failed to start process: %s"
            raise XStartError(msg % self)
        if not ok:
            msg = 'Failed to start X on display "%s" (xdpyinfo check failed, stderr:[%s]).'
            raise XStartTimeoutError(msg % (d, xdpyinfo.stderr))

    def _wait_for_pipe_text(self, rfd):
        s = ""
        start_time = time.time()
        while True:
            (rfd_changed_ls, _, _) = select.select([rfd], [], [], 0.1)
            if not self.is_alive():
                raise XStartError(
                    "%s program closed. command: %s stderr: %s"
                    % (self._program, self._command, self.stderr)
                )
            if rfd in rfd_changed_ls:
                c = os.read(rfd, 1)
                if c == b"\n":
                    break
                s += c.decode("ascii")

            # this timeout is for "eternal" hang. see #62
            if time.time() - start_time >= 600:  # = 10 minutes
                raise XStartTimeoutError(
                    "No reply from program %s. command:%s"
                    % (
                        self._program,
                        self._command,
                    )
                )
        return s

    def stop(self):
        """
        stop display

        :rtype: self
        """
        if not self._is_started:
            raise XStartError("stop() is called before start().")

        if self._reset_global_env:
            self._redirect_display(False)

        if self.is_alive():
            try:
                self._subproc.kill()
            except OSError as oserror:
                log.debug("exception in terminate:%s", oserror)

            self._subproc.wait()
            self._read_stdout_stderr()
        if self._use_xauth:
            self._clear_xauth()
        return self

    def _read_stdout_stderr(self):
        if self.stdout is None:
            (self.stdout, self.stderr) = self._subproc.communicate()

            log.debug("stdout=%s", self.stdout)
            log.debug("stderr=%s", self.stderr)

    def _setup_xauth(self):
        """
        Set up the Xauthority file and the XAUTHORITY environment variable.
        """
        handle, filename = tempfile.mkstemp(
            prefix="PyVirtualDisplay.", suffix=".Xauthority"
        )
        self._xauth_filename = filename
        os.close(handle)
        # Save old environment
        self._old_xauth = {}
        self._old_xauth["AUTHFILE"] = os.getenv("AUTHFILE")
        self._old_xauth["XAUTHORITY"] = os.getenv("XAUTHORITY")

        os.environ["AUTHFILE"] = os.environ["XAUTHORITY"] = filename
        cookie = xauth.generate_mcookie()
        xauth.call("add", self.new_display_var, ".", cookie)

    def _clear_xauth(self):
        """
        Clear the Xauthority file and restore the environment variables.
        """
        os.remove(self._xauth_filename)
        for varname in ["AUTHFILE", "XAUTHORITY"]:
            if self._old_xauth[varname] is None:
                del os.environ[varname]
            else:
                os.environ[varname] = self._old_xauth[varname]
        self._old_xauth = None

    def __enter__(self):
        """used by the :keyword:`with` statement"""
        self.start()
        return self

    def __exit__(self, *exc_info):
        """used by the :keyword:`with` statement"""
        self.stop()

    def is_alive(self):
        if not self._subproc:
            return False
        # return self.return_code is None
        rc = self._subproc.poll()
        if rc is not None:
            # proc exited
            self._read_stdout_stderr()
        return rc is None

    # @property
    # def return_code(self):
    #     if not self._subproc:
    #         return None
    #     rc = self._subproc.poll()
    #     if rc is not None:
    #         # proc exited
    #         self._read_stdout_stderr()
    #     return rc

    @property
    def pid(self):
        """
        PID (:attr:`subprocess.Popen.pid`)

        :rtype: int
        """
        if self._subproc:
            return self._subproc.pid



from time import sleep

import pytest
from tutil import has_xvnc, rfbport

from pyvirtualdisplay import Display
from pyvirtualdisplay.abstractdisplay import XStartError
from pyvirtualdisplay.xephyr import XephyrDisplay
from pyvirtualdisplay.xvfb import XvfbDisplay
from pyvirtualdisplay.xvnc import XvncDisplay


def test_virt():
    vd = Display()
    # assert vd.return_code is None
    assert not vd.is_alive()
    vd.start()
    # assert vd.return_code is None
    assert vd.is_alive()
    vd.stop()
    # assert vd.return_code == 0
    assert not vd.is_alive()

    vd = Display().start().stop()
    # assert vd.return_code == 0
    assert not vd.is_alive()


def test_nest():
    vd = Display().start()
    assert vd.is_alive()

    nd = Display(visible=True).start().stop()

    # assert nd.return_code == 0
    assert not nd.is_alive()

    vd.stop()
    assert not vd.is_alive()


def test_disp():
    vd = Display().start()
    assert vd.is_alive()

    # d = Display(visible=True).start().sleep(2).stop()
    # .assertEquals(d.return_code, 0)

    d = Display(visible=False).start().stop()
    # assert d.return_code == 0
    assert not d.is_alive()

    vd.stop()
    assert not vd.is_alive()


def test_repr_xvfb():
    display = Display()
    print(repr(display))

    display = Display(visible=False)
    print(repr(display))

    display = Display(backend="xvfb")
    print(repr(display))

    display = XvfbDisplay()
    print(repr(display))


if has_xvnc():

    def test_repr_xvnc():
        display = Display(backend="xvnc", rfbport=rfbport())
        print(repr(display))

        display = XvncDisplay()
        print(repr(display))


def test_repr_xephyr():
    display = Display(visible=True)
    print(repr(display))

    display = Display(backend="xephyr")
    print(repr(display))

    display = XephyrDisplay()
    print(repr(display))


def test_stop_nostart():
    with pytest.raises(XStartError):
        Display().stop()


def test_double_start():
    vd = Display()
    try:
        vd.start()
        with pytest.raises(XStartError):
            vd.start()
    finally:
        vd.stop()


def test_double_stop():
    vd = Display().start().stop()
    # assert vd.return_code == 0
    assert not vd.is_alive()
    vd.stop()
    # assert vd.return_code == 0
    assert not vd.is_alive()


def test_stop_terminated():
    vd = Display().start()
    assert vd.is_alive()
    vd._obj._subproc.kill()
    sleep(1)
    assert not vd.is_alive()
    vd.stop()
    # assert vd.return_code == 0
    assert not vd.is_alive()


def test_no_backend():
    with pytest.raises(ValueError):
        Display(backend="unknown")


def test_color_xvfb():
    with pytest.raises(XStartError):
        Display(color_depth=99).start().stop()
    Display(color_depth=16).start().stop()
    Display(color_depth=24).start().stop()
    Display(color_depth=8).start().stop()


def test_color_xephyr():
    with Display():
        # requested screen depth not supported, setting to match hosts
        Display(backend="xephyr", color_depth=99).start().stop()

        Display(backend="xephyr", color_depth=16).start().stop()
        Display(backend="xephyr", color_depth=24).start().stop()
        Display(backend="xephyr", color_depth=8).start().stop()


if has_xvnc():

    def test_color_xvnc():
        with pytest.raises(XStartError):
            with Display(backend="xvnc", color_depth=99, rfbport=rfbport()):
                pass
        with Display(backend="xvnc", color_depth=16, rfbport=rfbport()):
            pass
        with Display(backend="xvnc", color_depth=24, rfbport=rfbport()):
            pass
        # tigervnc no longer works 8-bit pseudocolors, 18.04 is OK
        # with Display(backend="xvnc", color_depth=8, rfbport=rfbport()):
        #     pass


def test_pid():
    with Display() as d:
        assert d.pid > 0
    with XvfbDisplay() as d:
        assert d.pid > 0


def test_bgcolor():
    Display(bgcolor="black").start().stop()
    Display(bgcolor="white").start().stop()
    with pytest.raises(KeyError):
        Display(bgcolor="green").start().stop()


def test_is_started():
    #     d = Display()
    #     assert not d._is_started
    #     d.start()
    #     assert d._is_started
    #     d.stop()
    #     assert d._is_started

    # with Display() as d:
    #     assert d._is_started
    # assert d._is_started

    with XvfbDisplay() as d:
        assert d._is_started
    assert d._is_started

    with Display():
        with XephyrDisplay() as d:
            assert d._is_started
        assert d._is_started

        # with XvncDisplay() as d:
        #     assert d._is_started
        # assert d._is_started


def test_extra_args():
    # Unrecognized option
    d = Display(extra_args=["willcrash"])
    with pytest.raises(XStartError):
        d.start()

    with Display():
        # -c                     turns off key-click
        with Display(visible=True, extra_args=["-c"]) as d:
            assert d.is_alive()
        assert not d.is_alive()

        with XephyrDisplay(extra_args=["-c"]) as d:
            assert d.is_alive()
        assert not d.is_alive()


def test_display():
    d = Display()
    assert d.display is None
    d.start()
    assert d.display >= 0

    d = XvfbDisplay()
    assert d.display is None
    d.start()
    assert d.display >= 0



from threading import Lock, Thread
from time import sleep

from easyprocess import EasyProcess

from pyvirtualdisplay import Display

disps = []
mutex = Lock()


def get_display(threadid, disp):
    dispnr = EasyProcess(["sh", "-c", "echo $DISPLAY"], env=disp.env()).call().stdout
    with mutex:
        disps.append((threadid, dispnr))


def func():
    with Display(manage_global_env=False) as disp:
        get_display(1, disp)
        sleep(2)
        get_display(1, disp)


def test_disp_var():
    t = Thread(target=func)
    t.start()
    sleep(1)

    with Display(manage_global_env=False) as disp:
        get_display(0, disp)
        sleep(2)
        get_display(0, disp)
    t.join()

    print(disps)

    assert disps[0][0] == 1
    assert disps[1][0] == 0
    assert disps[2][0] == 1
    assert disps[3][0] == 0

    # :1
    assert disps[0][1] == disps[2][1]

    # :0
    assert disps[1][1] == disps[3][1]

    assert disps[0][1] != disps[1][1]



#!/usr/bin/env python3
import os
from pathlib import Path
from time import sleep

import fabric
from entrypoint2 import entrypoint

import vagrant

# pip3 install fabric vncdotool python-vagrant entrypoint2

DIR = Path(__file__).parent


class Options:
    halt = True
    recreate = True
    destroy = False


def run_box(options, vagrantfile, cmds):
    env = os.environ
    if vagrantfile == "Vagrantfile":
        env["VAGRANT_VAGRANTFILE"] = str(DIR.parent.parent / vagrantfile)
        env["VAGRANT_DOTFILE_PATH"] = ""
    else:
        env["VAGRANT_VAGRANTFILE"] = str(DIR / vagrantfile)
        env["VAGRANT_DOTFILE_PATH"] = str(DIR / (".vagrant_" + vagrantfile))

    v = vagrant.Vagrant(env=env, quiet_stdout=False, quiet_stderr=False)
    status = v.status()
    state = status[0].state
    print(status)

    if options.destroy:
        v.halt(force=True)
        v.destroy()
        return

    if options.halt:
        v.halt()  # avoid screensaver

    if state == "not_created":
        # install programs in box
        v.up()
        # restart box
        v.halt()

    try:
        v.up()

        with fabric.Connection(
            v.user_hostname_port(),
            connect_kwargs={
                "key_filename": v.keyfile(),
            },
        ) as conn:
            with conn.cd("c:/vagrant" if options.win else "/vagrant"):
                if not options.win:
                    if options.osx:
                        freecmd = "top -l 1 -s 0 | grep PhysMem"
                    else:  # linux
                        freecmd = "free -h"
                    cmds = [freecmd, "env | sort"] + cmds + [freecmd]
                sleep(1)
                for cmd in cmds:
                    if options.recreate:
                        if "tox" in cmd:
                            cmd += " -r"
                    # hangs without pty=True
                    conn.run(cmd, echo=True, pty=True)
    finally:
        if options.halt:
            v.halt()


config = {
    "debian10": (
        "Vagrantfile.debian10.rb",
        ["tox -e py37"],
    ),
    "debian11": (
        "Vagrantfile.debian11.rb",
        ["tox -e py39"],
    ),
    "ubuntu2204": (
        "Vagrantfile.ubuntu2204.rb",
        ["tox -e py39"],
    ),
    "ubuntu2004": (
        "Vagrantfile",
        ["tox", "PYVIRTUALDISPLAY_DISPLAYFD=0 tox"],
    ),
    "ubuntu1804": (
        "Vagrantfile.ubuntu1804.rb",
        ["tox -e py36"],
    ),
    # "osx": (
    #     "Vagrantfile.osx.rb",
    #     [
    #         "bash --login -c 'python3 -m tox -e py3-osx'",
    #         # TODO: "bash --login -c 'PYVIRTUALDISPLAY_DISPLAYFD=0 python3 -m tox -e py3-osx'",
    #     ],
    # ),
}


@entrypoint
def main(boxes="all", fast=False, destroy=False):
    options = Options()
    options.halt = not fast
    options.recreate = not fast
    options.destroy = destroy

    if boxes == "all":
        boxes = list(config.keys())
    else:
        boxes = boxes.split(",")

    for k, v in config.items():
        name = k
        vagrantfile, cmds = v[0], v[1]
        if name in boxes:
            options.win = k.startswith("win")
            options.osx = k.startswith("osx")
            print("----->")
            print("----->")
            print("-----> %s %s %s" % (name, vagrantfile, cmds))
            print("----->")
            print("----->")
            try:
                run_box(options, vagrantfile, cmds)
            finally:
                print("<-----")
                print("<-----")
                print("<----- %s %s %s" % (name, vagrantfile, cmds))
                print("<-----")
                print("<-----")



import tempfile
from pathlib import Path

from tutil import has_xvnc, rfbport, worker
from vncdotool import api

from pyvirtualdisplay import Display
from pyvirtualdisplay.xvnc import XvncDisplay

if has_xvnc():

    def test_xvnc():
        with tempfile.TemporaryDirectory() as temp_dir:
            vnc_png = Path(temp_dir) / "vnc.png"
            password = "123456"
            passwd_file = Path(temp_dir) / "pwd.txt"
            vncpasswd_generated = b"\x49\x40\x15\xf9\xa3\x5e\x8b\x22"
            passwd_file.write_bytes(vncpasswd_generated)

            if worker() == 0:
                with Display(backend="xvnc"):
                    with api.connect("localhost:0") as client:
                        client.timeout = 1
                        client.captureScreen(vnc_png)
                with XvncDisplay():
                    with api.connect("localhost:0") as client:
                        client.timeout = 1
                        client.captureScreen(vnc_png)

            sconnect = "localhost:%s" % (rfbport() - 5900)
            with Display(backend="xvnc", rfbport=rfbport()):
                with api.connect(sconnect) as client:
                    client.timeout = 1
                    client.captureScreen(vnc_png)
            with XvncDisplay(rfbport=rfbport()):
                with api.connect(sconnect) as client:
                    client.timeout = 1
                    client.captureScreen(vnc_png)

            with Display(backend="xvnc", rfbport=rfbport(), rfbauth=passwd_file):
                with api.connect(sconnect, password=password) as client:
                    client.timeout = 1
                    client.captureScreen(vnc_png)
            with XvncDisplay(rfbport=rfbport(), rfbauth=passwd_file):
                with api.connect(sconnect, password=password) as client:
                    client.timeout = 1
                    client.captureScreen(vnc_png)



"Create screenshot of xmessage in background using 'smartdisplay' submodule"
from easyprocess import EasyProcess

from pyvirtualdisplay.smartdisplay import SmartDisplay

# 'SmartDisplay' instead of 'Display'
# It has 'waitgrab()' method.
# It has more dependencies than Display.
with SmartDisplay() as disp:
    with EasyProcess(["xmessage", "hello"]):
        # wait until something is displayed on the virtual display (polling method)
        # and then take a fullscreen screenshot
        # and then crop it. Background is black.
        img = disp.waitgrab()
img.save("xmessage.png")



import os.path

from setuptools import setup

NAME = "pyvirtualdisplay"

# get __version__
__version__ = None
exec(open(os.path.join(NAME, "about.py")).read())
VERSION = __version__

PYPI_NAME = "PyVirtualDisplay"
URL = "https://github.com/ponty/pyvirtualdisplay"
DESCRIPTION = "python wrapper for Xvfb, Xephyr and Xvnc"
LONG_DESCRIPTION = """pyvirtualdisplay is a python wrapper for Xvfb, Xephyr and Xvnc

Documentation: https://github.com/ponty/pyvirtualdisplay/tree/"""
LONG_DESCRIPTION += VERSION
PACKAGES = [
    NAME,
    NAME + ".examples",
]


classifiers = [
    # Get more strings from
    # http://www.python.org/pypi?%3Aaction=list_classifiers
    "License :: OSI Approved :: BSD License",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
]


setup(
    name=PYPI_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/x-rst",
    classifiers=classifiers,
    keywords="Xvfb Xephyr X wrapper",
    author="ponty",
    # author_email='',
    url=URL,
    license="BSD",
    packages=PACKAGES,
    # install_requires=install_requires,
    package_data={
        NAME: ["py.typed"],
    },
)



"Start virtual VNC server. Connect with: vncviewer localhost:5904"

from easyprocess import EasyProcess

from pyvirtualdisplay import Display

with Display(backend="xvnc", size=(100, 60), rfbport=5904) as disp:
    with EasyProcess(["xmessage", "hello"]) as proc:
        proc.wait()



import glob
import logging
import os
from time import sleep

from easyprocess import EasyProcess
from entrypoint2 import entrypoint

from pyvirtualdisplay.smartdisplay import SmartDisplay

# (cmd,grab,background)
commands = [
    ("python3 -m pyvirtualdisplay.examples.threadsafe", False, False),
    ("python3 -m pyvirtualdisplay.examples.screenshot", False, False),
    ("python3 -m pyvirtualdisplay.examples.lowres", True, True),
    ("python3 -m pyvirtualdisplay.examples.nested", True, True),
    ("python3 -m pyvirtualdisplay.examples.vncserver", False, True),
    ("vncviewer localhost:5904", True, True),
]


def screenshot(cmd, fname):
    logging.info("%s %s", cmd, fname)
    # fpath = "docs/_img/%s" % fname
    # if os.path.exists(fpath):
    #     os.remove(fpath)
    with SmartDisplay() as disp:
        with EasyProcess(cmd):
            img = disp.waitgrab()
            img.save(fname)


def empty_dir(dir):
    files = glob.glob(os.path.join(dir, "*"))
    for f in files:
        os.remove(f)


@entrypoint
def main():
    EasyProcess("killall Xvnc").call()
    gendir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "gen")
    logging.info("gendir: %s", gendir)
    os.makedirs(gendir, exist_ok=True)
    empty_dir(gendir)
    pls = []
    try:
        os.chdir("gen")
        for cmd, grab, bg in commands:
            sleep(1)

            with SmartDisplay() as disp:
                logging.info("======== cmd: %s", cmd)
                fname_base = cmd.replace(" ", "_")
                fname = fname_base + ".txt"
                # logging.info("cmd: %s", cmd)
                print("file name: %s" % fname)
                with open(fname, "w") as f:
                    f.write("$ " + cmd + "\n")
                    if bg:
                        p = EasyProcess(cmd).start()
                    else:
                        p = EasyProcess(cmd).call()
                        f.write(p.stdout)
                        f.write(p.stderr)
                    pls += [p]
                if grab:
                    png = fname_base + ".png"
                    sleep(1)
                    img = disp.waitgrab()
                    logging.info("saving %s", png)
                    img.save(png)
    finally:
        os.chdir("..")
        for p in reversed(pls):
            p.stop()
        EasyProcess("killall Xvnc").call()
    embedme = EasyProcess(["npx", "embedme", "../README.md"])
    embedme.call()
    print(embedme.stdout)
    assert embedme.return_code == 0
    assert "but file does not exist" not in embedme.stdout



import subprocess
import sys


def get_helptext(program):
    cmd = [program, "-help"]

    # py3.7+
    # p = subprocess.run(cmd, capture_output=True)
    # stderr = p.stderr

    # py3.6 also
    p = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=False,
    )
    _, stderr = p.communicate()

    helptext = stderr.decode("utf-8", "ignore")
    return helptext


def platform_is_osx():
    return sys.platform == "darwin"
